#include "bmp.h"
#include <stdint.h>
#include <iostream>     // std::cout
#include <fstream>      // std::ifstream

imageBMP::imageBMP(const char *name /** INPUT FILE NAME */ ){
   HEADER header;
   INFOHEADER infoheader;
   FILE *fptr;
   long size,m,cont;
   unsigned char *image;

   /* Open file */
   if ((fptr = fopen(name,"rb")) == NULL) {
      fprintf(stderr,"Unable to open BMP file \"%s\"\n",name);
      return;
   }

   /* Read and check the header */
   if (fread(&header.type,sizeof(unsigned short),1,fptr) != 1) {
       fprintf(stderr,"Image read failed\n");
       return;
   }

   /* Read and check the header */
   //fprintf(stderr,"ID is: %d, should be %d\n",header.type,'M'*256+'B');
   if (header.type != (int) 'M'*256+'B') {
     printf("Be careful, the file is not a bmp image file (magic number not correct)\n");
     //fprintf(stderr,"Be careful, the file is not a bmp image file (magic number not correct)\n");
     //fprintf(stderr,"  File : %s \n",name);
     //scanf("%d",&cont);
   }

   if (fread(&header.size,sizeof(unsigned int),1,fptr) != 1) {
               fprintf(stderr,"Image read failed\n");
               return;
   }

   //fprintf(stderr,"File size is %d bytes\n",header.size);
   if (fread(&header.reserved1,sizeof(unsigned short),1,fptr) != 1) {
               fprintf(stderr,"Image read failed\n");
               return;
   }

   if (fread(&header.reserved2,sizeof(unsigned short),1,fptr) != 1) {
               fprintf(stderr,"Image read failed\n");
               return;
   }
   //fprintf(stderr,"header.reserved1 = %d \n",header.reserved1);
   if (fread(&header.offset,sizeof(unsigned int),1,fptr) != 1) {
               fprintf(stderr,"Image read failed\n");
               return;
    }
    //fprintf(stderr,"header.reserved2 = %d \n",header.reserved2);
    //fprintf(stderr,"Offset to image data is %d bytes\n",header.offset);

   /* Read and check the information header */
   if (fread(&infoheader,sizeof(INFOHEADER),1,fptr) != 1) {
      fprintf(stderr,"Failed to read BMP info header\n");
      return;
   }
//   fprintf(stderr,"Header size = %d \n",infoheader.size);
//   fprintf(stderr,"Image size = %d x %d\n",infoheader.width,infoheader.height);
//   fprintf(stderr,"Number of colour planes is %d\n",infoheader.planes);
//   fprintf(stderr,"Bits per pixel is %d\n",infoheader.bits);
//   fprintf(stderr,"Compression type is %d\n",infoheader.compression);
//   fprintf(stderr,"Infoheader.imagesize is %d\n",infoheader.imagesize);
//   fprintf(stderr,"Infoheader.xresolution=%d infoheader.yresolution=%d \n",
//           infoheader.xresolution,infoheader.yresolution);
//   fprintf(stderr,"Number of colours is %d\n",infoheader.ncolours);
//   fprintf(stderr,"Number of required colours is %d\n",
//      infoheader.importantcolours);

   /* WE CHECK THE 24 bits */
   if(infoheader.bits!=24){
	   fprintf(stderr,"Be carefull, function only implemented for 24 bits colour format\n");
       //scanf("%d",&cont);
   }
   /* WE CHECK COMPRESSION */
   if(infoheader.compression!=0){
	   fprintf(stderr,"Sorry, function only implemented image without compression \n");
       return;
   }
   /* WE CHECK THE 24 bits */
   if(infoheader.ncolours!=0){
	   fprintf(stderr,"Be carefull, function only implemented for 24 bits colour format\n");
       //scanf("%d",&cont);
   }
   /* WE CHECK THE 24 bits */
   if(infoheader.importantcolours!=0){
	   fprintf(stderr,"Be carefull, function only implemented for 24 bits colour format\n");
       //scanf("%d",&cont);
   }


   /* WE FIT DIMENSIONS AND ALLOCATE MEMORY FOR THE IMAGE */
   width_=infoheader.width; height_=infoheader.height;
   size=width_*height_;
   RGB_.resize(size);

  /* WE STORE THE IMAGE INFORMATION IN image */
   image=(unsigned char*)malloc(3*size*sizeof(unsigned char));

   /* We read the image */
   if (fread(image,size*3,1,fptr) != 1) {
     fprintf(stderr,"Image read failed\n");
     free(image);
     return;
   }

   cont=0;
   for(m=0;m<size;m++){
       RGB_[m].R_=image[cont];
       RGB_[m].G_=image[cont+1];
       RGB_[m].B_=image[cont+2]; cont+=3;
   }
   fclose(fptr);

}


int imageBMP::write(const char *name){

   int width=4*(width_/4);
   int height=4*(height_/4);
   if(width==0 || height==0) return -1;
   if(width!=width_ || height!=height_){
     imageBMP I2(width,height);
     for(int x=0;x<width;x++){
       for(int y=0;y<height;y++){
         I2[y*width+x]=RGB_[y*width_+x];
       }
     }
     return I2.write(name);
   }

   HEADER header;
   INFOHEADER infoheader;
   FILE *fptr;
   long cont,size,m;
   unsigned char *image;

   /* Open file */
   if ((fptr = fopen(name,"wb")) == NULL) {
      fprintf(stderr,"Unable to open BMP file \"%s\"\n",name);
      return(-1);
   }

   /* Header Definition */
   header.type = (unsigned short int) 'M'*256+'B';
   header.size = (unsigned int) 3*width_*height_+54;
   header.reserved1 = 0;
   header.reserved2 = 0;
   header.offset = (unsigned int) 54;


  /* We write the header */
   if (fwrite(&header.type,sizeof(unsigned short),1,fptr) != 1) {
      fprintf(stderr,"Failed to write BMP info header\n");
      return(-1);
   }
   if (fwrite(&header.size,sizeof(unsigned int),1,fptr) != 1) {
      fprintf(stderr,"Failed to write BMP info header\n");
      return(-1);
   }
   if (fwrite(&header.reserved1,sizeof(unsigned short),1,fptr) != 1) {
      fprintf(stderr,"Failed to write BMP info header\n");
      return(-1);
   }
   if (fwrite(&header.reserved2,sizeof(unsigned short),1,fptr) != 1) {
      fprintf(stderr,"Failed to write BMP info header\n");
      return(-1);
   }
   if (fwrite(&header.offset,sizeof(unsigned int),1,fptr) != 1) {
      fprintf(stderr,"Failed to write BMP info header\n");
      return(-1);
   }


   /* INFOHEADER DEFINITION */
   infoheader.size = 40;
   infoheader.width = width_;
   infoheader.height = height_;
   infoheader.planes = (unsigned short int) 1;
   infoheader.bits = (unsigned short int) 24;
   infoheader.compression = (unsigned int) 0;
   infoheader.imagesize=3*width_*height_;
   infoheader.xresolution=2834;
   infoheader.yresolution=2834;
   infoheader.ncolours = (unsigned int) 0 ;
   infoheader.importantcolours = (unsigned int) 0;



   /* We write the infoheaderinformation */
   if (fwrite(&infoheader,sizeof(INFOHEADER),1,fptr) != 1) {
      fprintf(stderr,"Failed to write BMP info header\n");
      return(-1);
   }

   size=width_*height_;

   /* WE STORE THE IMAGE INFORMATION IN image */
   image=(unsigned char*)malloc(3*size*sizeof(unsigned char));
   //printf("size*3=%d\n",size*3);
   cont=0;
   for(m=0;m<size;m++){
       image[cont]=RGB_[m].R_;
       image[cont+1]=RGB_[m].G_;
       image[cont+2]=RGB_[m].B_;
       cont+=3;
   }

   /* We write the image */
   if (fwrite(image,size*3,1,fptr) != 1) {
     fprintf(stderr,"Image write failed\n");
     free(image);
     return(-1);
   }

   fclose(fptr);
   free(image);
   return(0);

}


/********************************************************************/
/* APPROXIMATION TO GAUSSIAN FILTER USING ALVAREZ-MAZORRA ALGORITHM */
/********************************************************************/
vector<float> ami_gauss_conv_alma(
vector<float> &rim2,  /* Original picture and output picture*/
float sigma_x,    /* standard deviation of the gaussian in the horizontal axis direction*/
float sigma_y,    /* standard deviation of the gaussian in the vertical axis direction */
float precision,  /* non-negative number to indicate the precisin of the convolution */
int width,int height)
{
   int m,M,cont;
   int i,j,nx1,ny1;

   vector<float> rim=rim2;

   vector<float> z;
   float l_x,v_x,l_y,v_y;
   int nx=height;
   int ny=width;
   int Nc_x=(int) (precision*sigma_x>1?precision*sigma_x:1);
   float t_x=sigma_x*sigma_x/(2*Nc_x);
   int Nc_y=(int) (precision*sigma_y>1?precision*sigma_y:1);
   float t_y=sigma_y*sigma_y/(2*Nc_y);
   int Nc_a=(Nc_x>Nc_y)?Nc_x:Nc_y;

   /* DEFINIMOS LOS PARAMETROS DEL ALGORITMO EN X E Y*/
   if(t_x>0){
     l_x=(1.+2.*t_x-sqrt((float) 4*t_x+1))/(2*t_x);
     v_x=l_x/t_x;
   }
   if(t_y>0){
     l_y=(1.+2.*t_y-sqrt((float) 4*t_y+1))/(2*t_y);
     v_y=l_y/t_y;
   }
   nx1=nx-1; ny1=ny-1;
   M=nx*ny;

   /* COGEMOS MEMORIA PARA EL VECTOR DE PASO */
   if(nx>ny)
	z.resize(nx);
   else
 	z.resize(ny);

   /* BUCLE PRINCIPAL DE LAS ITERACIONES */
   for(cont=0;cont<Nc_a;cont++){
     if(t_x>0 && cont<Nc_x){
       for(j=0;j<ny;j++){
         z[0]=rim[j]/(1-l_x);
         m=j;
         for(i=1;i<nx;i++){
           z[i]=rim[m+=ny]+l_x*z[i-1];
         }
         rim[m]=z[nx1]/(1-l_x);
         i=nx-1;
         while(i>0){
	       rim[m-ny]=(z[--i]+l_x*rim[m]);
           m-=ny;
         }
       }
	   m=0;
       while(m<M) rim[m++]*=v_x;
	 }
	 if(t_y>0 && cont<Nc_y){
       m=0;
       for(i=0;i<nx;i++){
         z[0]=rim[m]/(1-l_y);
         for(j=1;j<ny;j++){
           z[j]=rim[++m]+l_y*z[j-1];
         }
         rim[m]=z[ny1]/(1-l_y);
         j=ny1;
         while(j>0){
           --m;
	       rim[m]=z[--j]+l_y*rim[m+1];
         }
         m+=ny;
       }
       m=0;
       while(m<M) rim[m++]*=v_y;
     }
  }
  return(rim);
}

/******************************************************************************/
/*           FUNCTION TO COMPUTE HARRIS CORNER QUALITY VALUE                  */
/******************************************************************************/
vector<float> ami_harris_value_estimation(
imageBMP &I, /* ORIGINAL IMAGE */
float sigma) /* STANDARD DEVIATION FOR THE GAUSSIAN */
{
  int width=I.width();
  int height=I.height();
  int i,j,k,*flat,Ne=0;
  float ux,uy,c1,d1,coef1,coef2;
  int m,size=width*height;

  vector<float> harris_value(size); /* HARRIS QUALITY FACTOR */

  coef1=sqrt((float) 2.);
  coef2=0.25*(2.-coef1);
  coef1=0.5*(coef1-1);
  vector<float> image_x(size,0);
  vector<float> image_y(size,0);
  vector<float> image_xy(size,0);

  for(i=0;i<height;i++){
    for(j=0;j<width;j++){
      m=i*width+j;
      if(i==0 || i==(height-1) || j==0 || j==(width-1)){
	    ux=uy=0;
      }
      else{
	    c1= (float) I[m+width+1].R_-I[m-width-1].R_;
	    d1= (float) I[m-width+1].R_-I[m+width-1].R_;
	    uy=coef1*((float) I[m+width].R_-I[m-width].R_)+coef2*(c1-d1);
	    ux=-(coef1*( (float) I[m+1].R_-I[m-1].R_)+coef2*(c1+d1));

      }
      image_x[m]=ux*ux;
      image_y[m]=uy*uy;
      image_xy[m]=ux*uy;
    }
  }

  image_x=ami_gauss_conv_alma(image_x,sigma,sigma,5.,width,height);
  image_y=ami_gauss_conv_alma(image_y,sigma,sigma,5.,width,height);
  image_xy=ami_gauss_conv_alma(image_xy,sigma,sigma,5.,width,height);

  for(int m=0;m<size;m++){
      harris_value[m]=fabs(image_x[m]*image_y[m]-image_xy[m]*image_xy[m]-
	  0.04*(image_x[m]+image_y[m])*(image_x[m]+image_y[m]));
  }

  return(harris_value);
}


imageBMP imageBMP::symmetry(int number_of_sectors,int initial_sector,float center_x,float center_y,float angle0)
{
  float ami_PI=3.141617;
  imageBMP I2;
  int width,height;
  if(number_of_sectors==1){
    number_of_sectors=8;
    width=2*width_;
    height=2*height_;
    center_x=width_;
    center_y=height_;
    I2=imageBMP(width,height,color((unsigned char) 255,(unsigned char) 255,(unsigned char) 255));
    for(int x=0;x<width_;x++){
      for(int y=0;y<height_;y++){
        I2[(y+height_)*width+x+width_]=RGB_[y*width_+x];
        I2[(y+height_)*width+x]=RGB_[y*width_+x];
        I2[(y)*width+x+width_]=RGB_[y*width_+x];
        I2[(y)*width+x]=RGB_[y*width_+x];
      }
    }
  }
  else{
    I2=(*this);
    width=width_;
    height=height_;
  }



  for(int k=1;k<number_of_sectors;k++){
    printf("k=%d\n",k);
    float angle1=angle0+2.*ami_PI*(initial_sector+k)/number_of_sectors;
    float angle2=angle0+2.*ami_PI*(initial_sector+k+1)/number_of_sectors;
    printf("angle1=%lf\n angle2=%lf\n",angle1,angle2);
    float a1=sin(angle1); float b1=-cos(angle1); float c1=-a1*center_x-b1*center_y;
    float a2=sin(angle2); float b2=-cos(angle2); float c2=-a2*center_x-b2*center_y;
    for(int x=0;x<width;x++){
      for(int y=0;y<height;y++){
        if(a1*x+b1*y+c1<=0.1  && a2*x+b2*y+c2>=-0.1){
            int x0=round(x-2*a1*(a1*x+b1*y+c1));
            int y0=round(y-2*b1*(a1*x+b1*y+c1));
            if(0<=x0 && x0<width && 0<=y0 && y0<height)
              I2[x+y*width]=I2[x0+y0*width];
        }
      }
    }
  }

  return(I2);
}

imageBMP imageBMP::symmetry45()
{
  float ami_PI=3.141617;
  imageBMP I2=(*this);
  float angle0=0;
  float center_x=height_/2,center_y=height_/2;
  int initial_sector=2;

  int number_of_sectors=8;
  for(int k=1;k<4;k++){
    //printf("k=%d\n",k);
    float angle1=angle0+2.*ami_PI*(initial_sector+k)/number_of_sectors;
    float angle2=angle0+2.*ami_PI*(initial_sector+k+1)/number_of_sectors;
    //printf("angle1=%lf\n angle2=%lf\n",angle1,angle2);
    float a1=sin(angle1); float b1=-cos(angle1); float c1=-a1*center_x-b1*center_y;
    float a2=sin(angle2); float b2=-cos(angle2); float c2=-a2*center_x-b2*center_y;
    for(int x=0;x<width_;x++){
      for(int y=0;y<height_;y++){
        if(a1*x+b1*y+c1<=0.1  && a2*x+b2*y+c2>=-0.1){
            int x0=round(x-2*a1*(a1*x+b1*y+c1));
            int y0=round(y-2*b1*(a1*x+b1*y+c1));
            if(0<=x0 && x0<width_ && 0<=y0 && y0<height_)
              I2[x+y*width_]=I2[x0+y0*width_];
        }
      }
    }
  }

  for(int x=height_/2;x<width_;x++){
    for(int y=0;y<height_/2;y++){
      I2[x+y*width_]=I2[x+(height_-1-y)*width_];
    }
  }

   for(int x=width_/2;x<width_;x++){
    for(int y=0;y<height_;y++){
      I2[x+y*width_]=I2[width_-1-x+y*width_];
    }
  }

  return(I2);
}

void ami_date_name(char name[200])
{
  time_t timer;
  int m,j;
  char name2[200];
  time(&timer);
  m=((localtime(&timer))->tm_year+1900)*10000;
  m+=((localtime(&timer))->tm_mon+1)*100;
  m+=(localtime(&timer))->tm_mday;
  sprintf(name,"%d",m);

  m=1000000;
  m+=((localtime(&timer))->tm_hour)*10000;
  m+=((localtime(&timer))->tm_min)*100;
  m+=(localtime(&timer))->tm_sec;

  j=100000;
  while(j>=1){
	sprintf(name2,"%d",((int) ((int)(m/j))%10));
	j/=10;
    strcat(name,name2);
  }
  printf("name=%s\n",name);
}

vector < vector<int> > MorphoNeighborhood2D(int width){
  vector < vector<int> > MN;
  int k=0;

  // 1 0
  MN.resize(k+1);
  MN[k].push_back(1);
  MN[k++].push_back(-1);

  // 0 1
  MN.resize(k+1);
  MN[k].push_back(width);
  MN[k++].push_back(-width);

  // 1 1
  MN.resize(k+1);
  MN[k].push_back(1+width);
  MN[k++].push_back(-1-width);

  // 1 -1
  MN.resize(k+1);
  MN[k].push_back(1-width);
  MN[k++].push_back(-1+width);

  return(MN);

}

void morphological_dilation2D(imageBMP &I,int Niter){
//  printf("init morphological_dilation2D() ... ");
 // vector < vector<int> > MN=MorphoNeighborhood2D(I.width());

  // WE PERFORM ITERATIONS OF THE MORPHOLOGICAL EROSION
  for(int k=0;k<Niter;k++){
    // printf("k=%d\n",k);
    imageBMP I2=I;

    //WE COMPUTE DIFFERENTIAL OPERATOR AND UPDATE IMAGE I
    int height_1=I.height()-1;
    int width_1=I.width()-1;
    int width=I.width();
    for(int y=1;y<height_1;y++){
      int m0=y*width;
      for(int x=1;x<width_1;x++){
        int m=m0+x;
        I[m]=max(I2[m],I2[m+1],I2[m-1]);
        color c=max(I2[m],I2[m+width],I2[m-width]);
        I[m]=min(I[m],c);
        c=max(I2[m],I2[m+width+1],I2[m-width-1]);
        I[m]=min(I[m],c);
        c=max(I2[m],I2[m+width-1],I2[m-width+1]);
        I[m]=min(I[m],c);
      }
    }
  }
//  printf("end morphological_dilation2D()\n");
}

void morphological_erosion2D(imageBMP &I,int Niter){
//  printf("init morphological_dilation2D() ... ");
  vector < vector<int> > MN=MorphoNeighborhood2D(I.width());

  // WE PERFORM ITERATIONS OF THE MORPHOLOGICAL EROSION
  for(int k=0;k<Niter;k++){
    // printf("k=%d\n",k);
    imageBMP I2=I;

    //WE COMPUTE DIFFERENTIAL OPERATOR AND UPDATE IMAGE I
    int height_1=I.height()-1;
    int width_1=I.width()-1;
    int width=I.width();
    for(int y=1;y<height_1;y++){
      int m0=y*width;
      for(int x=1;x<width_1;x++){
        int m=m0+x;
        I[m]=min(I2[m],I2[m+1],I2[m-1]);
        color c=min(I2[m],I2[m+width],I2[m-width]);
        I[m]=max(I[m],c);
        c=min(I2[m],I2[m+width+1],I2[m-width-1]);
        I[m]=max(I[m],c);
        c=min(I2[m],I2[m+width-1],I2[m-width+1]);
        I[m]=max(I[m],c);
      }
    }
  }
//  printf("end morphological_dilation2D()\n");
}


imageBMP imageBMP::joint(imageBMP &I){
  if(height_==I.height()){
    imageBMP JointImage(width_+I.width(),height_);
    for(int y=0;y<height_;y++){
      for(int x=0;x<width_;x++){
        JointImage[y*JointImage.width()+x]=RGB_[y*width_+x];
      }
      for(int x=0;x<I.width();x++){
        JointImage[y*JointImage.width()+x+width_]=I[y*I.width()+x];
      }
    }
    return(JointImage);
  }

  int heightMax,heightMin;
  if(height_<I.height()){heightMax=I.height(); heightMin=height_;}
  else{ heightMax=height_; heightMin=I.height(); }
  imageBMP JointImage(width_+I.width(),heightMax,color((unsigned char)0,(unsigned char)0,(unsigned char)0));
  for(int y=0;y<heightMin;y++){
    for(int x=0;x<width_;x++){
      JointImage[y*JointImage.width()+x]=RGB_[y*width_+x];
    }
    for(int x=0;x<I.width();x++){
      JointImage[y*JointImage.width()+x+width_]=I[y*I.width()+x];
    }
  }
  return(JointImage);
}

imageBMP imageBMP::jointV(imageBMP &I){
  if(width_!=I.width()) return imageBMP();

  int size2d=width_*height_;
  imageBMP JointImage(width_,height_+I.height());
  for(int y=0;y<height_;y++){
    for(int x=0;x<width_;x++){
      JointImage[y*width_+x]=RGB_[y*width_+x];
    }
    for(int x=0;x<I.width();x++){
      JointImage[y*JointImage.width()+size2d+x]=I[y*width_+x];
    }
  }
  return(JointImage);
}

imageBMP imageBMP::zoom(const float &zoom){
  int widthN=width_*zoom;
  int heightN=height_*zoom;
  //printf("widthN=%d,heightN=%d\n",widthN,heightN);
  if(widthN<2 || heightN<2) return imageBMP();
  imageBMP iZ(widthN,heightN);
  for(int x=0;x<widthN;x++){
    for(int y=0;y<heightN;y++){
      float xf=x/zoom;
      float yf=y/zoom;
      int xi=xf;
      int yi=yf;
      if(xi>=width_ || yi>=height_){
        iZ[y*widthN+x]=color((unsigned char) 0,(unsigned char) 0,(unsigned char) 0);
        continue;
      }
      float xd=xf-xi;
      float yd=yf-yi;
      int xi1 = xi<(width_-1)?xi+1:xi;
      int yi1 = yi<(height_-1)?yi+1:yi;
      int l00=yi*width_+xi;
      int l10=yi*width_+xi1;
      int l01=yi1*width_+xi;
      int l11=yi1*width_+xi1;
      float c[3];
      //printf("xi=%d,y1=%d,xi1=%d,yi1=%d\n",xi,y1,xi1,yi1);
      c[0]=(float) (1.-xd)*(1.-yd)*RGB_[l00].R_+xd*(1.-yd)*RGB_[l10].R_+(1.-xd)*yd*RGB_[l01].R_+xd*yd*RGB_[l11].R_;
      c[1]=(float) (1.-xd)*(1.-yd)*RGB_[l00].G_+xd*(1.-yd)*RGB_[l10].G_+(1.-xd)*yd*RGB_[l01].G_+xd*yd*RGB_[l11].G_;
      c[2]=(float) (1.-xd)*(1.-yd)*RGB_[l00].B_+xd*(1.-yd)*RGB_[l10].B_+(1.-xd)*yd*RGB_[l01].B_+xd*yd*RGB_[l11].B_;
//      for(int k=0;k<3;k++){
//        c[k]=(float) (1.-xd)*(1.-yd)*RGB_[l00][k]+xd*(1.-yd)*RGB_[l10][k]+(1.-xd)*yd*RGB_[l01][k]+xd*yd*RGB_[l11][k];
//      }
      iZ[y*widthN+x]=color((unsigned char) c[0],(unsigned char) c[1],(unsigned char) c[2]);
    }
  }
  return iZ;
}

vector<float> zoom(vector<float> &If,int &width,int &height,float zoom){
  int widthN=width*zoom;
  int heightN=height*zoom;
  //printf("widthN=%d,heightN=%d\n",widthN,heightN);
  if(widthN<2 || heightN<2){
    width=0;
    height=0;
    return vector<float>();
  }
  vector<float> iZ(widthN*heightN);
  for(int x=0;x<widthN;x++){
    for(int y=0;y<heightN;y++){
      float xf=x/zoom;
      float yf=y/zoom;
      int xi=xf;
      int yi=yf;
      if(xi>=width || yi>=height){
        iZ[y*widthN+x]=0.;
        continue;
      }
      float xd=xf-xi;
      float yd=yf-yi;
      int xi1 = xi<(width-1)?xi+1:xi;
      int yi1 = yi<(height-1)?yi+1:yi;
      int l00=yi*width+xi;
      int l10=yi*width+xi1;
      int l01=yi1*width+xi;
      int l11=yi1*width+xi1;
      float c[3];
      //printf("xi=%d,y1=%d,xi1=%d,yi1=%d\n",xi,y1,xi1,yi1);
      iZ[y*widthN+x]=(1.-xd)*(1.-yd)*If[l00]+xd*(1.-yd)*If[l10]+(1.-xd)*yd*If[l01]+xd*yd*If[l11];
    }
  }
  width=widthN;
  height=heightN;
  return iZ;

}

///printf("WE ZOOM, BINARIZE AND FILTER INTENSITY IMAGE\n");
float zoom(vector<float> &If,int &width,int &height,float MinSqArea,float MaxSqArea,float sigma,float BinarizationSlope){
  float result=1.;;
  float sqArea=sqrt((float) width*height);
  if(sqArea>MaxSqArea){
    If=zoom(If,width,height,MaxSqArea/sqArea);
    sqArea=sqrt((float) width*height);
    result=sqArea/MinSqArea;
  }
  else if(sqArea<MinSqArea){
    If=zoom(If,width,height,MinSqArea/sqArea);
    sqArea=sqrt((float) width*height);
  }
  else{
    result=sqArea/MinSqArea;
  }

  If=ami_gauss_conv_alma(If,sigma,sigma,3.,width,height);
  for(int k=0;k<If.size();k++){
    If[k]=127.5+BinarizationSlope*(If[k]-127.5);
    if(If[k]<0.) If[k]=0.;
    else if(If[k]>255.) If[k]=255.;
  }
  return result;

}

/// SORT RANDOMLY
imageBMP imageBMP::sort(){
  double RAND_MAX2b=(double) RAND_MAX*RAND_MAX;
  imageBMP I2=*this;
  for(int k=0;k<RGB_.size();k++){
    int m=((double) rand()/RAND_MAX+rand()/RAND_MAX2b)*RGB_.size();
    m=m<RGB_.size()?m:RGB_.size()-1;
    color c0=I2.RGB_[m];
    I2.RGB_[m]=I2.RGB_[k];
    I2.RGB_[k]=c0;
  }
  return I2;
}

/// SORT HSV
imageBMP imageBMP::sortHSV(){
  imageBMP I2(width_,height_);
  vector<float> measure(RGB_.size());
  vector<int> index(RGB_.size());
  for(int k=0;k<RGB_.size();k++){
    float H,S,V;
    RGB_[k].rgb2hsv(H,S,V);
    measure[k]=4.*H+2.*S+V;
    index[k]=k;
  }

  for(int k=0;k<RGB_.size();k++){
    for(int m=k+1;m<RGB_.size();m++){
      if(measure[k]<measure[m]){
        float t0=measure[m];
        measure[m]=measure[k];
        measure[k]=t0;
        int i0=index[m];
        index[m]=index[k];
        index[k]=i0;
      }
    }
  }

  for(int k=0;k<RGB_.size();k++){
    I2.RGB_[k]=RGB_[index[k]];
  }

  return I2;
}


